// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "asset/AssetLoader.hpp"

namespace Dawn {
  class Game;

  class AssetManager final : public std::enable_shared_from_this<AssetManager> {
    private:
      std::weak_ptr<Game> game;
      std::vector<std::shared_ptr<AssetLoader>> pendingAssetLoaders;
      std::vector<std::weak_ptr<AssetLoader>> finishedAssetLoaders;

      /**
       * Returns an existing asset loader if it exists.
       * 
       * @param filename The filename of the asset to get.
       * @return The asset loader if it exists, otherwise nullptr.
       */
      template<class T>
      std::shared_ptr<T> getExisting(const std::string &filename) {
        for(auto &loader : pendingAssetLoaders) {
          if(loader->name != filename) continue;
          if(loader->getAssetType() != T::ASSET_TYPE) continue;
          return std::static_pointer_cast<T>(loader);
        }

        for(auto &wLoader : finishedAssetLoaders) {
          auto loader = wLoader.lock();
          if(!loader) continue;
          if(loader->name != filename) continue;
          if(loader->getAssetType() != T::ASSET_TYPE) continue;
          return std::static_pointer_cast<T>(loader);
        }
        
        return nullptr;
      }

      /**
       * Removes an existing asset loader if it exists.
       * 
       * @param filename The filename of the asset to remove.
       */
      void removeExisting(const std::string &filename);

    public:
      /**
       * Initializes this asset manager so it can begin accepting assets.
       * 
       * @param game Game context that this asset manager is attached to.
       */
      void init(const std::shared_ptr<Game> &game);

      /**
       * Updates the asset manager.
       */
      void update();

      /**
       * Returns whether the asset manager has loaded all of the currently
       * managed assets.
       * 
       * @return True if all assets have been loaded.
       */
      bool_t isEverythingLoaded();

      /**
       * Returns the asset loader for the given asset.
       * 
       * @param filename The filename of the asset to get.
       * @return The asset loader for the given asset.
       */
      template<class T>
      std::shared_ptr<T> get(const std::string &filename) {
        auto existing = this->getExisting<T>(filename);
        if(existing) return existing;

        std::shared_ptr<T> loader = std::make_shared<T>(
          shared_from_this(),
          filename
        );
        pendingAssetLoaders.push_back(
          std::static_pointer_cast<AssetLoader>(loader)
        );
        return loader;
      }

      /**
       * Returns the game context that this asset manager is attached to.
       * 
       * @return The game context.
       */
      std::shared_ptr<Game> getGame();

      /**
       * Removes the given asset loader from the asset manager, assumes that
       * nothing else needs to access it and any dangling shared_ptrs will have
       * to remain in memory.
       * 
       * @param loader The asset loader to remove.
       */
      void remove(const std::shared_ptr<AssetLoader> loader);

      /**
       * Dispose the asset manager, and all attached assets.
       */
      ~AssetManager();
  };
}