// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeLoader.hpp"
#include "assert/assert.hpp"

using namespace Dawn;

const std::string TrueTypeLoader::ASSET_TYPE = "ttf";

TrueTypeLoader::TrueTypeLoader(
  const std::shared_ptr<AssetManager> assetManager,
  const std::string name
) :
  AssetLoader(assetManager, name),
  loader(name)
{
  this->typetest = this->getAssetType();
  // Init the font.
  auto ret = FT_Init_FreeType(&fontLibrary);
  assertTrue(ret == 0, "Failed to initialize FreeType library.");
}

void TrueTypeLoader::updateSync() {
  if(state != TrueTypeLoaderState::ASYNC_DONE) return;
  state = TrueTypeLoaderState::SYNC_LOADING;

  // Init all the textures.
  auto it = textures.begin();
  while(it != textures.end()) {
    auto texture = it->second.lock();

    if(texture) {
      texture->setFace(face);
      it++;
      continue;
    }

    it = textures.erase(it);
  }

  // Done
  state = TrueTypeLoaderState::SYNC_DONE;
  this->loaded = true;
}

void TrueTypeLoader::updateAsync() {
  if(state != TrueTypeLoaderState::INITIAL) return;
  state = TrueTypeLoaderState::ASYNC_LOADING;

  // Load the data.
  this->loader.open();
  size_t size = loader.getSize();
  buffer = new uint8_t[size];

  // Read the data.
  size_t readSize = loader.read(buffer, size);
  assertTrue(readSize == size, "Failed to read all data from TrueTypeLoader.");

  // Init the font.
  auto ret = FT_New_Memory_Face(fontLibrary, buffer, size, 0, &face);
  assertTrue(ret == 0, "Failed to load font face.");

  // Now close the asset loader
  loader.close();
  state = TrueTypeLoaderState::ASYNC_DONE;
}

std::string TrueTypeLoader::getAssetType() const {
  return TrueTypeLoader::ASSET_TYPE;
}

std::shared_ptr<TrueTypeTexture> TrueTypeLoader::getTexture(
  const uint32_t fontSize
) {
  // Check if we have the texture already and it hasn't gone stale.
  auto it = textures.find(fontSize);
  if(it != textures.end()) {
    if(!it->second.expired()) return it->second.lock();
    textures.erase(it);
  }

  // Create the texture.
  auto texture = std::make_shared<TrueTypeTexture>(fontSize);
  textures[fontSize] = texture;
  if(this->loaded) texture->setFace(face);
  return texture;
}

TrueTypeLoader::~TrueTypeLoader() {
  if(
    this->state == TrueTypeLoaderState::SYNC_DONE ||
    this->state == TrueTypeLoaderState::SYNC_LOADING ||
    this->state == TrueTypeLoaderState::ASYNC_DONE
  ) {
    FT_Done_Face(face);
  }

  FT_Done_FreeType(fontLibrary);

  if(buffer != nullptr) {
    delete[] buffer;
    buffer = nullptr;
  }
}