// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "LoaderForSceneItems.hpp"
#include "asset/loader/TextureLoader.hpp"
#include "asset/loader/scene/PrefabLoader.hpp"

using namespace Dawn;

LoaderForSceneItems::LoaderForSceneItems(
  const std::shared_ptr<AssetManager> assetManager,
  const std::string name
) :
  AssetLoader(assetManager, name),
  ctx(std::make_shared<SceneLoadContext>())
{
}

void LoaderForSceneItems::setupDependencies() {
  assertNotNull(this->jsonLoader, "JSON Loader is NULL?");
  assertNotNull(ctx, "SceneLoadContext is NULL?");
  assertTrue(this->jsonLoader->loaded, "JSON loader not loaded?");
  
  // Begin loading dependencies.
  auto &data = this->jsonLoader->data;
  if(data.contains("assets")) {
    for(auto &asset : data["assets"].items()) {
      auto &assetName = asset.key();
      auto &assetData = asset.value();
      assertTrue(assetData.contains("type"), "Asset missing type");
      assertTrue(assetData.contains("path"), "Asset missing path");
      auto type = assetData["type"].get<std::string>();
      auto path = assetData["path"].get<std::string>();

      // Is this asset already named?
      if(ctx->assets.find(assetName) != ctx->assets.end()) {
        // Check if path and type already the same.
        auto &existing = ctx->assets[assetName];
        assertTrue(
          existing->name == path && existing->getAssetType() == type,
          "Asset already exists with different path or type: %s",
          assetName.c_str()
        );
        continue;
      }

      std::shared_ptr<AssetLoader> loader;
      if(type == "texture") {
        loader = getAssetManager()->get<TextureLoader>(path);

      } else if(type == "json") {
        loader = getAssetManager()->get<JSONLoader>(path);

      } else if(type == "prefab") {
        auto prefabLoader = getAssetManager()->get<PrefabLoader>(path);
        prefabLoader->ctx->parent = ctx;
        loader = prefabLoader;

      } else {
        assertUnreachable("Unknown asset type: %s", type.c_str());
      }

      ctx->assets[assetName] = loader;
    }
  }
}

LoaderForSceneItems::~LoaderForSceneItems() {
  jsonLoader = nullptr;
}