// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "assert/assert.hpp"

namespace Dawn {
  class Scene;
  class SceneItem;
  class SceneComponent;
  class AssetLoader;
  class SceneLoadContext;

  class PrefabLoader;
  class LoaderForSceneItems;
  class SceneLoader;

  class SceneLoadContext {
    private:
      std::weak_ptr<SceneLoadContext> parent;
      std::unordered_map<std::string, std::shared_ptr<SceneItem>> items;
      std::unordered_map<std::string, std::shared_ptr<SceneComponent>> components;
      std::unordered_map<std::string, std::shared_ptr<AssetLoader>> assets;
      std::shared_ptr<Scene> currentScene;
      std::shared_ptr<SceneItem> currentItem;
      std::shared_ptr<SceneComponent> currentComponent;

    public:
      json data;

      /**
       * Gets an asset from the context.
       * 
       * @param j Name of the asset to get.
       * @return Asset from the context.
       */
      template<class T>
      std::shared_ptr<T> getAsset(const std::string &j) const {
        auto it = assets.find(j);
        if(it == assets.end()) {
          auto parent = this->parent.lock();
          assertNotNull(parent, "Couldn't find asset %s", j.c_str());
          return parent->getAsset<T>(j);
        }
        auto asset = std::dynamic_pointer_cast<T>(it->second);
        assertNotNull(asset, "Asset is not of the correct type.");
        return asset;
      }

      /**
       * Gets an item from the context.
       * 
       * @param j Name of the item to get.
       * @return Item from the context.
       */
      std::shared_ptr<SceneItem> getItem(const std::string &j);

      /**
       * Gets a component from the context.
       * 
       * @param j Name of the component to get.
       * @return Component from the context.
       */
      template<class T>
      std::shared_ptr<T> getComponent(const std::string &j) const {
        auto it = components.find(j);
        if(it == components.end()) {
          auto parent = this->parent.lock();
          assertNotNull(parent, "Couldn't find component.");
          return parent->getComponent<T>(j);
        }
        auto cmp = std::dynamic_pointer_cast<T>(it->second);
        assertNotNull(cmp, "Component is not of the correct type.");
        return cmp;
      }

      friend class PrefabLoader;
      friend class LoaderForSceneItems;
      friend class SceneLoader;
      friend class SceneItem;
  };
}