// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "asset/loader/scene/SceneLoadContext.hpp"

namespace Dawn {
  class Game;
  class Scene;
  class SceneItem;

  enum class SceneComponentState : flag_t {
    INITIAL = FLAG(0),
    INITIALIZED = FLAG(1),
    DISPOSED = FLAG(2)
  };

  class SceneComponent : std::enable_shared_from_this<SceneComponent> {
    private:
      std::weak_ptr<SceneItem> item;
      flag_t state = (flag_t)SceneComponentState::INITIAL;

    protected:
      std::vector<std::function<void()>> events;
      
      /**
       * Custom component listener that is invoked when the component is meant
       * to initialize.
       */
      virtual void onInit() = 0;

      /**
       * Custom component listener that is invoked when the component is meant
       * to dispose.
       */
      virtual void onDispose() = 0;

    public:
      std::string name;

      /**
       * Initializes this scene component.
       * 
       * @param item Scene item that this component belongs to.
       */
      void init(const std::shared_ptr<SceneItem> item);

      /**
       * Disposes this scene component.
       */
      void dispose();

      /**
       * Returns whether this scene component is initialized.
       * 
       * @return Whether this scene component is initialized.
       */
      bool_t isInitialized();

      /**
       * Returns the scene item that this scene component belongs to.
       * 
       * @return Reference to the scene item that this component belongs to.
       */
      std::shared_ptr<SceneItem> getItem();

      /**
       * Returns the scene that this scene component belongs to.
       * 
       * @return Reference to the scene that this component belongs to.
       */
      std::shared_ptr<Scene> getScene();

      /**
       * Returns the game that this scene component belongs to.
       * 
       * @return Reference to the game that this component belongs to.
       */
      std::shared_ptr<Game> getGame();

      /**
       * Load data from a JSON object. This is typically done during a scene
       * load.
       * 
       * @param json JSON Data that this object needs to load.
       */
      virtual void load(std::shared_ptr<SceneLoadContext> context);

      /**
       * Disposes this scene component.
       */
      virtual ~SceneComponent();
  };
}