// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "assert/assert.hpp"
#include "scene/item/SceneItemTransform.hpp"
#include "scene/SceneItem.hpp"

using namespace Dawn;

SceneItemTransform::SceneItemTransform() :
  transformLocal(1.0f),
  transformWorld(1.0f)
{
  this->updateLocalValuesFromLocalTransform();
}

void SceneItemTransform::updateLocalValuesFromLocalTransform() {
  glm::vec3 skew;
  glm::vec4 perspective;
  glm::decompose(
    this->transformLocal,
    this->localScale,
    this->localRotation,
    this->localPosition,
    skew,
    perspective
  );
}

void SceneItemTransform::updateLocalTransformFromWorldTransform() {
  auto parent = this->getParent();
  if(!parent) {
    this->transformLocal = this->transformWorld;
  } else {
    this->transformLocal = (
      glm::inverse(parent->transformWorld) * this->transformWorld
    );
  }
  this->updateLocalValuesFromLocalTransform();
}

void SceneItemTransform::updateChildrenWorldTransforms() {
  std::for_each(
    this->children.begin(),
    this->children.end(),
    [](std::weak_ptr<SceneItem> &item) {
      if(auto child = item.lock()) child->updateWorldTransformFromParent();
    }
  );
}

void SceneItemTransform::updateWorldTransformFromParent() {
  auto parent = this->getParent();
  if(!parent) {
    this->transformWorld = this->transformLocal;
  } else {
    this->transformWorld = (
      parent->transformWorld * this->transformLocal
    );
  }
  this->updateLocalValuesFromLocalTransform();
  this->updateChildrenWorldTransforms();
}

void SceneItemTransform::updateLocalTransformFromLocalValues() {
  this->transformLocal = glm::translate(
    glm::mat4(1.0f),
    this->localPosition
  ) * glm::mat4_cast(this->localRotation) * glm::scale(
    glm::mat4(1.0f),
    this->localScale
  );
  this->updateWorldTransformFromLocalTransform();
}

void SceneItemTransform::updateWorldTransformFromLocalTransform() {
  auto parent = this->getParent();
  if(!parent) {
    this->transformWorld = this->transformLocal;
  } else {
    this->transformWorld = (
      parent->transformWorld * this->transformLocal
    );
  }
  this->updateChildrenWorldTransforms();
}

std::shared_ptr<SceneItem> SceneItemTransform::getParent() {
  return parent.lock();
}

std::vector<std::shared_ptr<SceneItem>> SceneItemTransform::getChildren() {
  std::vector<std::shared_ptr<SceneItem>> children;
  std::for_each(
    this->children.begin(),
    this->children.end(),
    [&children](std::weak_ptr<SceneItem> &item) {
      if(auto child = item.lock()) children.push_back(child);
    }
  );
  return children;
}

glm::mat4 SceneItemTransform::getLocalTransform() {
  return this->transformLocal;
}

glm::mat4 SceneItemTransform::getWorldTransform() {
  return this->transformWorld;
}

glm::vec3 SceneItemTransform::getLocalPosition() {
  return this->localPosition;
}

glm::vec3 SceneItemTransform::getLocalScale() {
  return this->localScale;
}

glm::quat SceneItemTransform::getLocalRotation() {
  return this->localRotation;
}

void SceneItemTransform::setLocalTransform(const glm::mat4 transform) {
  this->transformLocal = transform;
  this->updateLocalValuesFromLocalTransform();
  this->updateWorldTransformFromLocalTransform();
}

void SceneItemTransform::setWorldTransform(const glm::mat4 transform) {
  this->transformWorld = transform;
  this->updateLocalTransformFromWorldTransform();
  this->updateChildrenWorldTransforms();
}

void SceneItemTransform::setLocalPosition(const glm::vec3 position) {
  this->localPosition = position;
  this->updateLocalTransformFromLocalValues();
}

void SceneItemTransform::setLocalScale(const glm::vec3 scale) {
  this->localScale = scale;
  this->updateLocalTransformFromLocalValues();
}

void SceneItemTransform::setLocalRotation(const glm::quat rotation) {
  this->localRotation = rotation;
  this->updateLocalTransformFromLocalValues();
}

bool_t SceneItemTransform::isChildOf(std::shared_ptr<SceneItem> item) {
  assertNotNull(item, "Cannot check if child of null item.");
  auto parent = this->getParent();
  while(parent) {
    if(parent == item) return true;
    parent = parent->getParent();
  }
  return false;
}

void SceneItemTransform::lookAt(
  const glm::vec3 position,
  const glm::vec3 target,
  const glm::vec3 up
) {
  this->setWorldTransform(glm::lookAt(position, target, up));
}

float_t SceneItemTransform::lookAtPixelPerfect(
  const glm::vec3 &position,
  const glm::vec3 &look,
  const float_t &viewportHeight,
  const float_t &fov,
  const float_t &scale
) {
  float_t z = (
    tanf((glm::radians(180.0f) - fov) / 2.0f) *
    (viewportHeight/2.0f)
  ) / scale;
  this->lookAt(
    glm::vec3(position.x, position.y, position.z + z),
    look,
    glm::vec3(0, 1, 0)
  );
  return z;
}

SceneItemTransform::~SceneItemTransform() {
  std::for_each(
    this->children.begin(),
    this->children.end(),
    [](std::weak_ptr<SceneItem> &item) {
      if(auto child = item.lock()) {
        child->parent.reset();
        child->updateWorldTransformFromParent();
      }
    }
  );
}