// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "ShaderStage.hpp"
#include "assert/assert.hpp"
#include "assert/assertgl.hpp"

using namespace Dawn;

void ShaderStage::init(
  const SlangStage &stage,
  const std::string &code
) {
  assertTrue(this->id == -1, "ShaderStage already initialized?");

  // Determine GL Shader type from slang shader type.
  switch(stage) {
    case SlangStage::SLANG_STAGE_VERTEX:
      shaderType = GL_VERTEX_SHADER;
      break;

    case SlangStage::SLANG_STAGE_FRAGMENT:
      shaderType = GL_FRAGMENT_SHADER;
      break;

    default:
      assertUnreachable("Unknown Slang Shader type\n");
      break;
  }

  // Initialize the shader.
  this->id = glCreateShader(shaderType);
  assertNoGLError();

  // Compile
  const char_t* cSource = code.c_str();
  glShaderSource(this->id, 1, &cSource, nullptr);
  assertNoGLError();

  glCompileShader(this->id);
  assertNoGLError();

  // Validate shader compiled successfully.
  GLint status;
  glGetShaderiv(this->id, GL_COMPILE_STATUS, &status);
  assertNoGLError();

  if(!status) {
    // Failed to compile
    GLint logLength;
    glGetShaderiv(this->id, GL_INFO_LOG_LENGTH, &logLength);
    assertNoGLError();

    GLchar *log = new GLchar[logLength];
    glGetShaderInfoLog(this->id, logLength, NULL, log);
    assertNoGLError();
    assertUnreachable("Failed to compile shader stage %i:\n%s", shaderType, log);
  }
}

ShaderStage::~ShaderStage() {
  if(this->id != -1) {
    glDeleteShader(this->id);
    assertNoGLError();
  }
}